@@ -3,9 +3,154 @@ import REFPROPConnector.Support.constants as constants
 from .Support.resources.file_handler import RP_EXEC
 from abc import ABC, abstractmethod
 from sty import fg, bg, ef, rs
+from copy import deepcopy
 import warnings, sty
 
+
 GLOBALCounter = 0
+CODES_TO_BE_ITERATED = [
+
+    "HQ", "SQ", "EQ",
+    "QH", "QS", "QE"
+
+]
+
+
+class QualityIteration:
+
+    def __init__(self, RPHandler, str_in: str, str_out: str, a: float, b: float):
+
+        self.RPHandler = RPHandler
+        self.str_in = str_in
+
+        self.__identify_Q(a, b)
+        self.__identify_T_range()
+        self.__iterate_quality()
+
+        self.result = self.RPHandler.calculate("TQ", str_out, self.T_limits[0], self.q_value)
+
+    def __identify_Q(self, a, b):
+
+        self.other_var = self.str_in.strip("Q")
+
+        if self.str_in[0] == "Q":
+
+            self.q_value = a
+            self.var_value = b
+
+        else:
+
+            self.q_value = b
+            self.var_value = a
+
+    def __identify_T_range(self):
+
+        min_T = self.RPHandler.calculate("EOSMIN", "T", 0., 0.)
+
+        try:
+
+            triple_T = self.RPHandler.calculate("TRIP", "T", 0., 0.)
+
+        except:
+
+            triple_T = -1000000
+
+        self.min_T = max(min_T, triple_T)
+        self.T_limits = [min_T, self.RPHandler.TC]
+
+    def __iterate_quality(self):
+
+        self.value_limits = [
+
+            self.__calculate_var(self.T_limits[0]),
+            self.__calculate_var(self.T_limits[1])
+
+        ]
+
+        if not (self.value_limits[0] * self.value_limits[1] < 0):
+            raise Exception("Unable to perform {} flash".format(self.str_in))
+
+        counter = 0
+        while abs(self.T_limits[0] - self.T_limits[1]) > 10**-3 or counter > 30:
+            counter += 1
+            self.__quality_iteration_step()
+
+    def __quality_iteration_step(self):
+
+        # Bisection Calculation
+        T_bis = (self.T_limits[1] + self.T_limits[0]) / 2
+        bis_dict = self.__get_quality_iteration_step_result(T_bis)
+        dict_list = [bis_dict]
+
+        # Secant Calculation (if possible)
+        T_val = self.T_limits
+        res_val = self.value_limits
+        T_sec = T_val[0] - res_val[0] / (res_val[1] - res_val[0]) * (T_val[1] - T_val[0])
+
+        if self.min_T < T_sec < self.RPHandler.TC:
+
+            sec_dict = self.__get_quality_iteration_step_result(T_sec)
+            dict_list.append(sec_dict)
+
+            # if possible check the interval between T_sec and T_bis
+            if bis_dict["var_new"] * sec_dict["var_new"] < 0:
+                dict_list.append({
+
+                    "new_limits": {
+
+                        "T": [bis_dict["T_new"], sec_dict["T_new"]],
+                        "var": [bis_dict["var_new"], sec_dict["var_new"]]
+
+                    },
+                    "dT": abs(bis_dict["T_new"] - sec_dict["T_new"])
+
+                })
+
+        # take the minimum dt interval
+        min_dict = bis_dict
+        for res_dict in dict_list:
+
+            if min_dict["dT"] > res_dict["dT"]:
+                min_dict = res_dict
+
+        self.T_limits = min_dict["new_limits"]["T"]
+        self.value_limits = min_dict["new_limits"]["var"]
+
+    def __get_quality_iteration_step_result(self, new_T):
+
+        T_limits_new = deepcopy(self.T_limits)
+        value_limits_new = deepcopy(self.value_limits)
+        new_var = self.__calculate_var(new_T)
+
+        if new_var * self.value_limits[1] < 0:
+
+            T_limits_new[0] = new_T
+            value_limits_new[0] = new_var
+
+        else:
+
+            T_limits_new[1] = new_T
+            value_limits_new[1] = new_var
+
+        return {
+
+            "T_new": new_T,
+            "var_new": new_var,
+            "new_limits": {
+
+                "T": T_limits_new,
+                "var": value_limits_new
+
+            },
+            "dT": abs(T_limits_new[1] - T_limits_new[0])
+
+        }
+
+    def __calculate_var(self, T_value):
+
+        value = self.RPHandler.calculate("TQ", self.other_var, T_value, self.q_value)
+        error = value - self.var_value
+        return error
 
 
 class RefPropHandler:
@@ -62,6 +207,24 @@ class RefPropHandler:
 
     def calculate(self, str_in: str, str_out: str, a: float, b: float):
 
+        global GLOBALCounter
+
+        if str_in in CODES_TO_BE_ITERATED:
+
+            qi = QualityIteration(self, str_in, str_out, a, b)
+            return qi.result
+
+        if str_out == "Q":
+
+            if self.unit_is_mass_based:
+
+                str_out = "QMASS"
+
+            else:
+
+                str_out = "QMOLE"
+
+        GLOBALCounter += 1
         self.refprop.SETFLUIDSdll('*'.join(self.fluids))
         return self.refprop.REFPROP1dll(str_in, str_out, self.SI, 1, a, b, self.composition).c
 
@@ -124,6 +287,11 @@ class RefPropHandler:
 
         self.set_reference_state(old_unit_system=old_unit_system)
 
+    @property
+    def unit_is_mass_based(self):
+
+        return not ("mol" in self.return_units("s"))
+
     @property
     def T_0_in_K(self):
 
@@ -340,7 +508,7 @@ class AbstractThermodynamicPoint(ABC):
 
                 self.calculate_on_need_variables.extend(self.state_var_list)
 
-    def __calculate(self):
+    def __update_variables(self):
 
         not_none_variables = self.not_none_variables
         self.calculated_variables = not_none_variables
@@ -372,23 +540,35 @@ class AbstractThermodynamicPoint(ABC):
 
     def __calculate_direct(self, REFPROP_CODE):
 
-        global GLOBALCounter
-
         if self.calculation_ready:
 
-            GLOBALCounter += 1
-
             not_none_variables = self.not_none_variables
             input_str = not_none_variables[0].refprop_name + not_none_variables[1].refprop_name
 
-            value = self.RPHandler.calculate(
+            if input_str not in CODES_TO_BE_ITERATED:
 
-                input_str,
-                REFPROP_CODE,
-                not_none_variables[0].value,
-                not_none_variables[1].value
+                value = self.RPHandler.calculate(
 
-            )
+                    input_str,
+                    REFPROP_CODE,
+                    not_none_variables[0].value,
+                    not_none_variables[1].value
+
+                )
+
+            else:
+
+                P_value = self.RPHandler.calculate(
+
+                    input_str,
+                    "P",
+                    not_none_variables[0].value,
+                    not_none_variables[1].value
+
+                )
+
+                self.set_variable("P", P_value)
+                value = self.__calculate_direct(REFPROP_CODE)
 
             return value
 
@@ -416,7 +596,7 @@ class AbstractThermodynamicPoint(ABC):
 
         if self.calculation_ready:
 
-            self.__calculate()
+            self.__update_variables()
 
     def get_variable(self, variable_name: str):
 
